# Copyright (c) 2019 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Migrate from schema version 0 to version 1.

This migration is a null-migration that serves as a template
for future migrations and testing purposes.
"""
import os

from signac._vendor import configobj

# A minimal v1 config.
_CFG = """
schema_version = string(default='0')
project = string()
workspace_dir = string(default='workspace')
"""


def _load_config_v1(root_directory):
    config_fn = os.path.join(root_directory, "signac.rc")
    if not os.path.isfile(config_fn):
        raise RuntimeError(
            f"The directory {root_directory} does not contain a config file."
        )
    cfg = configobj.ConfigObj(config_fn, configspec=_CFG.split("\n"))
    validator = configobj.validate.Validator()
    if cfg.validate(validator) is not True:
        raise RuntimeError(
            "This project's config file is not compatible with signac's v1 schema."
        )
    return cfg


def _migrate_v0_to_v1(root_directory):
    """Migrate from schema version 0 to version 1."""
    pass



# Copyright (c) 2022 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Implement class for indexing signac Projects."""

import json
import logging
import operator
import re
from math import isclose
from numbers import Number

from ._utility import _nested_dicts_to_dotted_keys, _to_hashable
from .errors import InvalidKeyError

logger = logging.getLogger(__name__)

_PRIMARY_KEY = "_id"

_INDEX_OPERATORS = (
    "$eq",
    "$gt",
    "$gte",
    "$lt",
    "$lte",
    "$ne",
    "$in",
    "$nin",
    "$regex",
    "$type",
    "$where",
    "$near",
)

_TYPES = {
    "int": int,
    "float": float,
    "bool": bool,
    "str": str,
    "list": tuple,
    "null": type(None),
}


class _DictPlaceholder:
    pass


class _float(float):
    # Numerical objects of either integer or float type, that share the same numerical value,
    # but not the same type, are distinguished within a search index, but considered equal within
    # Python. We manipulate the hash value, to enable the storage of both an int and a float
    # that share the same numerical value within a search index (dict).

    # There is no risk of accidentally equating ints and floats with different values, since the
    # hash equality is only a necessary, not a sufficient condition for equality.
    def __hash__(self):
        return super().__hash__() + 1


class _TypedSetDefaultDict(dict):
    """Dictionary that is guaranteed to store differently typed values separately.

    This is necessary, because the hash value of integers with float type is identical
    to the same integer as int type, which means they cannot be stored separately in a
    standard dict.

    """

    def keys(self):
        for key in dict.keys(self):
            yield float(key) if type(key) is _float else key

    __iter__ = keys

    def items(self):
        for key, value in dict.items(self):
            yield float(key) if type(key) is _float else key, value

    def __missing__(self, key):
        value = set()
        dict.__setitem__(self, key, value)
        return value

    def __getitem__(self, key):
        return dict.__getitem__(self, _float(key) if type(key) is float else key)

    def __setitem__(self, key, value):
        return dict.__setitem__(self, _float(key) if type(key) is float else key, value)

    def __delitem__(self, key):
        dict.__delitem__(self, _float(key) if type(key) is float else key)

    def get(self, key, default=None):
        """Get the value for given key.

        Parameters
        ----------
        key : str, float
            The key to get the value.
        default : object, optional
            Default value if type of key is not float (Default value = None).

        Returns
        -------
        The value for given key.

        """
        return dict.get(self, _float(key) if type(key) is float else key, default)


def _find_with_index_operator(index, op, argument):
    """Find index for given operator and argument.

    Parameters
    ----------
    index : dict
        Index for the operator.
    op : str
        logical operator.
    argument :
        Dependent on the choice of logical operator argument (op).
        For better understanding have a look at :meth:`~_SearchIndexer.find`.

    Returns
    -------
    set
        Index for given operator and argument.

    Raises
    ------
    ValueError
        When unknown argument is given for $type operator (When the operator is $type).

    """
    if op == "$in":

        def op(value, argument):
            return value in argument

    elif op == "$nin":

        def op(value, argument):
            return value not in argument

    elif op == "$regex":

        def op(value, argument):
            if isinstance(value, str):
                return re.search(argument, value)
            else:
                return False

    elif op == "$type":

        def op(value, argument):
            if argument in _TYPES:
                t = _TYPES[argument]
            else:
                raise ValueError(f"Unknown argument for $type operator: '{argument}'.")
            return isinstance(value, t)

    elif op == "$where":

        def op(value, argument):
            return eval(argument)(value)

    elif op == "$near":
        rel_tol, abs_tol = 1e-9, 0.0  # default values
        if isinstance(argument, (list, tuple)):
            if len(argument) == 1:
                argument = argument[0]
            elif len(argument) == 2:
                argument, rel_tol = argument
            elif len(argument) == 3:
                argument, rel_tol, abs_tol = argument
            else:
                err_msg = (
                    "The argument of the $near operator must be a float or a list of floats with "
                    "length 1, 2, or 3."
                )
                raise ValueError(err_msg)
        argument = float(argument)
        rel_tol = float(rel_tol)
        abs_tol = float(abs_tol)

        def op(value, argument):
            return isclose(value, argument, rel_tol=rel_tol, abs_tol=abs_tol)

    else:
        op = getattr(operator, {"$gte": "$ge", "$lte": "$le"}.get(op, op)[1:])
    matches = set()
    for value in index:
        if op(value, argument):
            matches.update(index[value])
    return matches


def _check_logical_operator_argument(op, argument):
    """Check arguments for the logical-operator.

    Parameters
    ----------
    op : str
        logical-operator.
    argument : list
        list of arguments for logical-operator.

    Raises
    ------
    ValueError
        The argument of logical-operator is not a list or is an empty list.

    """
    if not isinstance(argument, list):
        raise ValueError(f"The argument of logical-operator '{op}' must be a list!")
    if not len(argument):
        raise ValueError(f"The argument of logical-operator '{op}' cannot be empty!")


class _SearchIndexer(dict):
    """A searchable index of dicts.

    The _SearchIndexer class is a :class:`dict` that maps from ids to
    :class:`dict`s. The :class:`dict`s stored as values can be searched by
    their contained keys and values, returning ids for the values matching the
    provided query. The query syntax is based on MongoDB, though this class
    does not aim to match the API of MongoDB's Collection class.

    The dictionary values may be nested (may contain other dicts or lists), but
    have two restrictions. First, the data must be JSON-encodable. Second, the
    keys in the dictionary may not contain dots (``.``).

    For example, suppose we are given dictionaries of member data containing a
    `name` key and an `age` key along with unique identifiers acting as a
    primary key for each member. We can find the name of all members that are
    age 32 like this:

    .. code-block:: python

        members = _SearchIndexer({
            '0': {'name': 'John',  'age': 32},
            '1': {'name': 'Alice', 'age': 28},
            '2': {'name': 'Kevin', 'age': 32},
            # ...
        })

        for member_id in members.find({'age': 32}):
            print(member_id)  # prints 0 and 2

    Because this class inherits from :class:`dict`, it can be constructed in
    any of the same ways as a :class:`dict`, like ``_SearchIndexer(**kwargs)``,
    ``_SearchIndexer(mapping, **kwargs)``, or
    ``_SearchIndexer(iterable, **kwargs)``.

    """

    def build_index(self, key):
        """Build index for a given key.

        This is a highly performance critical code path.

        Parameters
        ----------
        key : str
            The key on which the index is built.

        Returns
        -------
        :class:`~_TypedSetDefaultDict`
            Index for key.

        Raises
        ------
        :class:`~signac.errors.InvalidKeyError`
            The dict contains invalid keys.

        """
        logger.debug(f"Building index for key '{key}'...")
        nodes = key.split(".")
        index = _TypedSetDefaultDict()

        for _id, doc in self.items():
            try:
                v = doc
                # Recursively access nested values from dotted keys.
                for n in nodes:
                    v = v[n]
            except (KeyError, TypeError):
                pass
            else:
                # `isinstance(instance, cls)` is typically faster than `type(instance) is cls`
                # when the answer is True, but it is slower when it is False. Since we
                # expect lists and dicts to occur infrequently here, we optimize for the
                # False path using the `type` based check.
                if type(v) is list:
                    index[_to_hashable(v)].add(_id)
                elif type(v) is dict:
                    index[_DictPlaceholder].add(_id)
                else:
                    index[v].add(_id)

            # Raise an exception if the original key is present and has dots.
            if len(nodes) > 1 and key in doc:
                raise InvalidKeyError(
                    "Keys with dots ('.') are invalid.\n\n"
                    "See https://signac.readthedocs.io/en/latest/"
                    "recipes.html#apply-document-wide-changes "
                    "for a recipe on how to replace dots in existing keys."
                )
        logger.debug(f"Built index for key '{key}'.")
        return index

    def _find_expression(self, key, value):
        """Find ids of dicts with keys matching a value expression.

        Parameters
        ----------
        key : str
            The dict key to match.
        value
            The value expression to match.

        Returns
        -------
        set
            The ids of dicts matching the value expression.

        Raises
        ------
        KeyError
            An invalid operator was given.
        ValueError
            The value is not bool for '$exists' operator or not a
            supported type for '$type' operator.

        """
        logger.debug(f"Find ids matching expression '{key}: {value}'.")
        if "$" in key:
            if key.count("$") > 1:
                raise KeyError(f"Invalid operator expression '{key}'.")
            nodes = key.split(".")
            op = nodes[-1]
            if not op.startswith("$"):
                raise KeyError(f"Invalid operator placement '{key}'.")
            key = ".".join(nodes[:-1])
            if op in _INDEX_OPERATORS:
                index = self.build_index(key)
                return _find_with_index_operator(index, op, value)
            elif op == "$exists":
                if not isinstance(value, bool):
                    raise ValueError(
                        "The value of the '$exists' operator must be boolean."
                    )
                index = self.build_index(key)
                match = {elem for elems in index.values() for elem in elems}
                return match if value else set(self).difference(match)
            else:
                raise KeyError(f"Unknown operator '{op}'.")
        else:
            index = self.build_index(key)
            # Check to see if 'value' is a floating point type but an
            # integer value (e.g., 4.0), and search for both the int and float
            # values. This allows the user to find statepoints that have
            # integer-valued keys that are stored as floating point types.
            # Note that this both cases: 1) user searches for an int and hopes
            # to find values that are stored as integer-valued floats and 2) user
            # searches for a integer-valued float and hopes to find ints.
            # This way, both `signac find x 4.0` and `signac find x 4` would
            # return jobs where `sp.x` is stored as either 4.0 or 4.
            if isinstance(value, Number) and float(value).is_integer():
                result_float = index.get(_float(value), set())
                result_int = index.get(int(value), set())
                return result_int.union(result_float)
            else:
                return index.get(value, set())

    def _find_result(self, expr):
        """Find ids of dicts matching a dict of filter expressions.

        Parameters
        ----------
        expr : dict
            The filter of expressions to match.

        Returns
        -------
        set
            A set of ids of dicts that match the given filter.

        """
        if not expr:
            # Empty expression yields all ids.
            return set(self)

        result_ids = None

        def reduce_results(match):
            """Reduce the results by intersection of matches.

            Parameters
            ----------
            match : set
                match for the given expression.

            """
            nonlocal result_ids
            if result_ids is None:  # First match
                result_ids = match
            else:  # Update previous match
                result_ids = result_ids.intersection(match)

        # Check if filter contains primary key, in which case we can
        # immediately reduce the result.
        _id = expr.pop(_PRIMARY_KEY, None)
        if _id is not None:
            reduce_results({_id} if _id in self else set())

        # Extract all logical-operator expressions for now.
        or_expressions = expr.pop("$or", None)
        and_expressions = expr.pop("$and", None)
        not_expression = expr.pop("$not", None)

        # Reduce the result based on the remaining non-logical expression:
        for key, value in _nested_dicts_to_dotted_keys(expr):
            reduce_results(self._find_expression(key, value))
            if not result_ids:
                # No matches, so exit early.
                return set()

        # Reduce the result based on the logical-operator expressions:
        if not_expression is not None:
            not_match = self._find_result(not_expression)
            reduce_results(set(self).difference(not_match))
            if not result_ids:
                # No matches, so exit early.
                return set()

        if and_expressions is not None:
            _check_logical_operator_argument("$and", and_expressions)
            for expr_ in and_expressions:
                reduce_results(self._find_result(expr_))
                if not result_ids:
                    # No matches, so exit early.
                    return set()

        if or_expressions is not None:
            _check_logical_operator_argument("$or", or_expressions)
            or_results = set()
            for expr_ in or_expressions:
                or_results.update(self._find_result(expr_))
            reduce_results(or_results)

        return result_ids

    def find(self, filter_=None):
        """Find ids of dicts matching a dict of filter expressions.

        This function normalizes the filter argument and then attempts to
        build a set of ids matching the given key-value queries.
        For each key that is queried, an internal index is built and then
        searched.

        The results are a set of ids, where each id is the value of the
        primary key of a dict that matches the given filter.

        The find() method uses the following optimizations:

            1. If the filter is None, a set of all ids is returned.
            2. The filter is processed key by key. Once the set of matches is
               empty it is immediately returned.

        Parameters
        ----------
        filter_ : dict, optional
            The filter of expressions to match (Default value = None).

        Returns
        -------
        set
            A set of ids of dicts that match the given filter.

        Raises
        ------
        ValueError
            When the filter argument is invalid.

        """
        if not filter_:
            return set(self)

        filter_ = json.loads(json.dumps(filter_))  # Normalize
        if not isinstance(filter_, dict):
            raise ValueError(f"Invalid filter: {filter_}")
        return self._find_result(filter_)



# Copyright (c) 2019 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Handle migrations of signac schema versions."""

import os

from filelock import FileLock

from .._utility import _print_err
from ..version import SCHEMA_VERSION, __version__
from .v0_to_v1 import _load_config_v1, _migrate_v0_to_v1
from .v1_to_v2 import _load_config_v2, _migrate_v1_to_v2

FN_MIGRATION_LOCKFILE = ".SIGNAC_PROJECT_MIGRATION_LOCK"


# Config loaders must be functions with the signature
# def config_loader(root_directory: str) -> MutableMapping
# When a new schema version is introduced, a corresponding loader only needs to
# be added if the old loader will no longer function.  This dictionary must
# contain all unique loaders for schema versions that are supported as starting
# points for migration. The resulting MutableMapping config objects must be
# writeable, i.e. it must be possible to persist in-memory changes from these
# objects to the underlying config files.
_CONFIG_LOADERS = {
    1: _load_config_v1,
    2: _load_config_v2,
}


_MIGRATIONS = {
    (0, 1): _migrate_v0_to_v1,
    (1, 2): _migrate_v1_to_v2,
}

_VERSION_LIST = list(reversed(sorted(_CONFIG_LOADERS.keys())))


def _get_config_schema_version(root_directory, version_guess):
    # Try loading the schema using the loader corresponding to the expected
    # version if it has a configured loader.
    versions = _VERSION_LIST
    if version_guess in _CONFIG_LOADERS:
        versions = [version_guess] + versions
    for guess in versions:
        try:
            # Note: We could consider using a different component as the key
            # for _CONFIG_LOADERS, but since this is an internal detail it's
            # not terribly consequential.
            config = _CONFIG_LOADERS[guess](root_directory)
            break
        except Exception:
            # The load failed, go to the next
            pass
    else:
        raise RuntimeError("Unable to load config file.")
    try:
        return int(config["schema_version"])
    except KeyError:
        # The default schema version is version 0.
        return 0


def _collect_migrations(root_directory):
    schema_version = int(SCHEMA_VERSION)

    current_schema_version = _get_config_schema_version(root_directory, schema_version)
    if current_schema_version > schema_version:
        # Project config schema version is newer and therefore not supported.
        raise RuntimeError(
            "The signac schema version used by this project is "
            f"{current_schema_version}, but signac {__version__} only "
            f"supports up to schema version {SCHEMA_VERSION}. Try updating "
            "signac."
        )

    guess = current_schema_version
    while _get_config_schema_version(root_directory, guess) < schema_version:
        for (origin, destination), migration in _MIGRATIONS.items():
            if origin == _get_config_schema_version(root_directory, guess):
                yield (origin, destination), migration
                guess = destination
                break
        else:
            raise RuntimeError(
                "The signac schema version used by this project is "
                f"{_get_config_schema_version(root_directory, guess)}, but "
                f"signac {__version__} uses schema version {schema_version} "
                "and does not know how to migrate."
            )


def apply_migrations(root_directory):
    """Apply migrations to a project.

    This function identifies and performs all the necessary schema migrations
    to bring a project up to date with the current schema version of signac.
    The calling code does not require prior knowledge of the schema version of
    the project, and the function is idempotent when applied to projects that
    already have an up-to-date schema.

    Parameters
    ----------
    root_directory : str
        The path to the project to migrate.
    """
    try:
        lock = FileLock(os.path.join(root_directory, FN_MIGRATION_LOCKFILE))
        with lock:
            for (origin, destination), migrate in _collect_migrations(root_directory):
                try:
                    _print_err(
                        f"Applying migration for version {origin} to {destination}... ",
                        end="",
                    )
                    migrate(root_directory)
                except Exception as e:
                    raise RuntimeError(
                        f"Failed to apply migration {destination}."
                    ) from e
                else:
                    config = _CONFIG_LOADERS[destination](root_directory)
                    config["schema_version"] = destination
                    config.write()
                    _print_err("OK")
    finally:
        try:
            os.unlink(lock.lock_file)
        except FileNotFoundError:
            pass


__all__ = [
    "apply_migrations",
]



# Copyright (c) 2019 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import os
import pickle
from tempfile import TemporaryDirectory

import pytest

from signac.h5store import H5StoreManager

try:
    import h5py  # noqa

    H5PY = True
except ImportError:
    H5PY = False


@pytest.mark.skipif(not H5PY, reason="test requires the h5py package")
class TestH5StoreManager:
    @pytest.fixture(autouse=True)
    def setUp(self, request):
        self._tmp_dir = TemporaryDirectory(prefix="h5store_")
        request.addfinalizer(self._tmp_dir.cleanup)
        self.store = H5StoreManager(prefix=self._tmp_dir.name)
        with open(os.path.join(self._tmp_dir.name, "other_file.txt"), "w") as file:
            file.write(r"blank\n")

    def test_repr(self):
        assert eval(repr(self.store)) == self.store

    def test_str(self):
        assert eval(str(self.store)) == self.store

    def test_set(self):
        assert len(self.store) == 0
        assert "test" not in self.store
        for value in ("", [], {}):
            with pytest.raises(ValueError):
                self.store["test"] = value
        for value in (True, 0, 0.0, 1, 1.0, None):
            with pytest.raises(TypeError):
                self.store["test"] = value
        for value in "abc":
            with pytest.raises(ValueError):
                self.store["test"] = value

        # Assigning a dictionary is the intended use case
        self.store["test"] = dict(foo=True)
        assert len(self.store) == 1
        assert "test" in self.store

    def test_set_iterable(self):
        assert len(self.store) == 0
        assert "test" not in self.store
        self.store["test"] = list(dict(foo=True).items())
        assert len(self.store) == 1
        assert "test" in self.store

    def test_set_get(self):
        assert len(self.store) == 0
        assert "test" not in self.store
        self.store["test"]["foo"] = "bar"
        assert "test" in self.store
        assert len(self.store) == 1
        assert "foo" in self.store["test"]

    def test_del(self):
        assert len(self.store) == 0
        assert "test" not in self.store
        self.store["test"]["foo"] = "bar"
        assert "test" in self.store
        assert len(self.store) == 1
        assert "foo" in self.store["test"]
        with pytest.raises(KeyError):
            del self.store["invalid"]
        del self.store["test"]
        assert len(self.store) == 0
        assert "test" not in self.store

    def test_iteration(self):
        keys = ["foo", "bar", "baz"]
        for key in keys:
            self.store[key] = dict(test=True)
        assert list(sorted(keys)) == list(sorted(self.store))
        assert list(sorted(keys)) == list(sorted(self.store.keys()))

    def test_contains(self):
        keys = ["foo", "bar", "baz"]
        for key in keys:
            assert key not in self.store
        for key in keys:
            self.store[key] = dict(test=True)
        for key in keys:
            assert key in self.store

    def test_pickle(self):
        assert pickle.loads(pickle.dumps(self.store)) == self.store



# Copyright (c) 2017 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Parses signac config files."""

import logging
import os

from ._vendor.configobj import ConfigObj, ConfigObjError
from ._vendor.configobj.validate import Validator
from .errors import ConfigError

logger = logging.getLogger(__name__)

PROJECT_CONFIG_FN = os.path.join(".signac", "config")
USER_CONFIG_FN = os.path.expanduser(os.path.join("~", ".signacrc"))
_CFG = """
schema_version = string(default='1')
"""


def _get_project_config_fn(path):
    return os.path.abspath(os.path.join(path, PROJECT_CONFIG_FN))


def _raise_if_older_schema(root):
    """Raise if an older schema version is detected at the search path.

    Parameters
    ----------
    root : str
        Directory to check schema for.

    Raises
    ------
    IncompatibleSchemaVersion
        If the project uses an older schema version that requires migration.
    """
    from .errors import IncompatibleSchemaVersion
    from .migration import _get_config_schema_version
    from .version import SCHEMA_VERSION, __version__

    try:
        schema_version = _get_config_schema_version(root, int(SCHEMA_VERSION))
        assert schema_version != int(SCHEMA_VERSION), (
            "Migration schema loader succeeded in loading a config file "
            "where normal loader failed. Do you have config files for multiple "
            "schemas? Otherwise, this indicates an internal "
            "error. Please contact the signac developers."
        )
        raise IncompatibleSchemaVersion(
            "Detected signac project using schema version "
            f"{schema_version}, but signac {__version__} requires "
            f"schema version {SCHEMA_VERSION}. Try running python -m "
            "signac migrate."
        )
    except RuntimeError:
        pass


def _locate_config_dir(search_path):
    """Locates directory containing a signac configuration file in a directory hierarchy.

    Parameters
    ----------
    search_path : str
        Starting path to search.

    Returns
    -------
    str or None
        The directory containing the configuration file if one is found, otherwise None.
    """
    orig_search_path = search_path
    search_path = os.path.abspath(search_path)
    while True:
        if os.path.isfile(_get_project_config_fn(search_path)):
            return search_path
        if (up := os.path.dirname(search_path)) == search_path:
            break
        else:
            search_path = up

    logger.debug(
        "Reached filesystem root, no config found. Checking whether a "
        "project created with an older signac schema may be found."
    )

    search_path = os.path.abspath(orig_search_path)
    while True:
        _raise_if_older_schema(search_path)
        if (up := os.path.dirname(search_path)) == search_path:
            logger.debug("Reached filesystem root, no config found.")
            return None
        else:
            search_path = up


class _Config(ConfigObj):
    """Manages configuration for a signac project."""

    def verify(self, *, preserve_errors=False):
        """Validate the contents of this configuration."""
        return super().validate(Validator(), preserve_errors=preserve_errors)


def _read_config_file(filename):
    logger.debug(f"Reading config file '{filename}'.")
    try:
        config = _Config(filename, configspec=_CFG.split("\n"))
    except (OSError, ConfigObjError) as error:
        raise ConfigError(f"Failed to read configuration file '{filename}':\n{error}")
    verification = config.verify()
    # config.verify() returns True if everything succeeded, but if the
    # validation failed it will return a dictionary of invalid results. We
    # cannot simply check for a truthy value here since a non-empty dict will
    # evaluate to True.
    if verification is not True:
        raise ConfigError(
            f"Config file '{os.path.abspath(filename)}' may contain invalid values."
        )
    return config


def _load_config(path=None):
    """Load configuration from a project directory.

    Parameters
    ----------
    path : str
        The project path to pull project-local configuration data from.

    Returns
    -------
    :class:`_Config`
        The composite configuration including both project-local and global
        config data if requested. Note that because this config is a composite,
        modifications to the returned value will not be reflected in the files.
    """
    if path is None:
        path = os.getcwd()
    config = _Config(configspec=_CFG.split("\n"))

    # Add in any global or user config files. For now this only finds user-specific
    # files, but it could be updated in the future to support e.g. system-wide config files.
    for fn in (USER_CONFIG_FN,):
        if os.path.isfile(fn):
            config.merge(_read_config_file(fn))

    if os.path.isfile(_get_project_config_fn(path)):
        config.merge(_read_config_file(_get_project_config_fn(path)))
    return config



# Copyright (c) 2017 The Regents of the University of Michigan.
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Job class defined here."""

import errno
import hashlib
import json
import logging
import os
import shutil
from copy import deepcopy
from threading import RLock
from types import MappingProxyType
from typing import FrozenSet

from synced_collections.backends.collection_json import (
    BufferedJSONAttrDict,
    JSONAttrDict,
    json_attr_dict_validator,
)
from synced_collections.errors import KeyTypeError
from synced_collections.utils import SyncedCollectionJSONEncoder

from ._utility import _mkdir_p
from .errors import DestinationExistsError, JobsCorruptedError
from .h5store import H5StoreManager
from .sync import sync_jobs

logger = logging.getLogger(__name__)


def calc_id(statepoint):
    """Calculate and return a hash value for the given statepoint.

    The hash is computed as an MD5 checksum of the input data. The input data
    is first encoded as JSON, with dictionary keys sorted to ensure the hash
    is reproducible.

    Parameters
    ----------
    statepoint : dict
        A JSON-encodable mapping.

    Returns
    -------
    str
        Encoded hash in hexadecimal format.
    """
    blob = json.dumps(statepoint, cls=SyncedCollectionJSONEncoder, sort_keys=True)
    m = hashlib.md5()
    m.update(blob.encode())
    return m.hexdigest()


# Note: All children of _StatePointDict will be of its parent type because they
# share a backend and the SyncedCollection registry parses the classes in order
# of registration. _If_ we need more control over this, that process can be
# exposed more thoroughly and registration can be made explicit rather than
# implicit, but for now the existing behavior works fine.
class _StatePointDict(JSONAttrDict):
    """A JSON-backed dictionary for storing job state points.

    There are three principal reasons for extending the base JSONAttrDict:
        1. Saving needs to trigger a job directory migration, and
        2. State points are assumed to not support external modification, so
           they never need to load from disk _except_ the very first time a job
           is opened by id and the state point is not present in the cache.
        3. It must be possible to load and/or save on demand during tasks like
           job directory migrations.
    """

    _PROTECTED_KEYS: FrozenSet[str] = JSONAttrDict._PROTECTED_KEYS.union(("_jobs",))
    _all_validators = (json_attr_dict_validator,)

    def __init__(
        self,
        jobs=None,
        filename=None,
        write_concern=False,
        data=None,
        parent=None,
        *args,
        **kwargs,
    ):
        # Multiple Python Job objects can share a single `_StatePointDict`
        # instance because they are shallow copies referring to the same data
        # on disk. We need to store these jobs in a shared list here so that
        # shallow copies can point to the same place and trigger each other to
        # update. This does not apply to independently created Job objects,
        # even if they refer to the same disk data; this only applies to
        # explicit shallow copies and unpickled objects within a session.
        self._jobs = list(jobs)
        super().__init__(
            filename=filename,
            write_concern=write_concern,
            data=data,
            parent=parent,
            *args,
            **kwargs,
        )

    def _load(self):
        # State points never load from disk automatically. They are either
        # initialized with provided data (e.g. from the state point cache), or
        # they load from disk the first time state point data is requested for
        # a Job opened by id (in which case the state point must first be
        # validated manually).
        pass

    def _save(self):
        # State point modification triggers job migration for all jobs sharing
        # this state point (shallow copies of a single job).
        new_id = calc_id(self)

        # All elements of the job list are shallow copies of each other, so any
        # one of them is representative.
        job = next(iter(self._jobs))
        old_id = job._id
        if old_id == new_id:
            return

        tmp_statepoint_file = self.filename + "~"
        should_init = False
        try:
            # Move the state point to an intermediate location as a backup.
            os.replace(self.filename, tmp_statepoint_file)
            try:
                new_workspace = os.path.join(job._project.workspace, new_id)
                os.replace(job.path, new_workspace)
            except OSError as error:
                os.replace(tmp_statepoint_file, self.filename)  # rollback
                if error.errno in (errno.EEXIST, errno.ENOTEMPTY, errno.EACCES):
                    raise DestinationExistsError(new_id)
                else:
                    raise
            else:
                should_init = True
        except OSError as error:
            # The most likely reason we got here is because the state point
            # file move failed due to the job not being initialized so the file
            # doesn't exist, which is OK.
            if error.errno != errno.ENOENT:
                raise

        # Update each job instance.
        for job in self._jobs:
            job._id = new_id
            job._initialize_lazy_properties()

        # Remove the temporary state point file if it was created. Have to do it
        # here so that we can get the updated job state point filename and set
        # it to self.filename below.
        try:
            os.remove(job._statepoint_filename + "~")
        except OSError as error:
            if error.errno != errno.ENOENT:
                raise

        self.filename = job._statepoint_filename

        if should_init:
            # Only initializing one job assumes that all changes in init are
            # changes reflected in the underlying resource (the JSON file).
            # This assumption is currently valid because all in-memory
            # attributes are loaded lazily (and are handled by the call to
            # _initialize_lazy_properties above), except for the key defining
            # property of the job id (which is also updated above). If init
            # ever changes to making modifications to the job object, we may
            # need to call it for all jobs.
            job.init()

        logger.info(f"Moved '{old_id}' -> '{new_id}'.")

    def save(self, force=False):
        """Trigger a save to disk.

        Unlike normal JSONAttrDict objects, this class requires the ability to save
        on command. Moreover, this save must be conditional on whether or not a
        file is present to allow the user to observe state points in corrupted
        data spaces and attempt to recover.

        Parameters
        ----------
        force : bool
            If True, save even if the file is present on disk.
        """
        try:
            # Open the file for writing only if it does not exist yet.
            if force or not os.path.isfile(self.filename):
                super()._save()
        except Exception as error:
            if not isinstance(error, OSError) or error.errno not in (
                errno.EEXIST,
                errno.EACCES,
            ):
                # Attempt to delete the file on error, to prevent corruption.
                # OSErrors that are EEXIST or EACCES don't need to delete the file.
                try:
                    os.remove(self.filename)
                except Exception:  # ignore all errors here
                    pass
                raise

    def load(self, job_id):
        """Trigger a load from disk.

        Unlike normal JSONAttrDict objects, this class requires the ability to
        load on command. These loads typically occur when the state point
        must be validated against the data on disk; at all other times, the
        in-memory data is assumed to be accurate to avoid unnecessary I/O.

        Parameters
        ----------
        job_id : str
            Job id used to validate contents on disk.

        Returns
        -------
        data : dict
            Dictionary of state point data.

        Raises
        ------
        :class:`~signac.errors.JobsCorruptedError`
            If the data on disk is invalid or its hash does not match the job
            id.

        """
        try:
            data = self._load_from_resource()
        except json.JSONDecodeError:
            raise JobsCorruptedError([job_id])

        if calc_id(data) != job_id:
            raise JobsCorruptedError([job_id])

        with self._suspend_sync:
            self._update(data, _validate=False)

        return data


class Job:
    """The job instance is a handle to the data of a unique state point.

    Application developers should not directly instantiate this class, but
    use :meth:`~signac.Project.open_job` instead.

    Jobs can be opened by ``statepoint`` or ``id_``. If both values are
    provided, it is the user's responsibility to ensure that the values
    correspond. Set ``directory_known`` to ``True`` when the job directory
    is known to exist - this skips some expensive isdir checks.

    Parameters
    ----------
    project : :class:`~signac.Project`
        Project handle.
    statepoint : dict, optional
        State point for the job. (Default value = None)
    id_ : str, optional
        The job identifier. (Default value = None)
    directory_known : bool, optional
        Set to true when the job directory is known to exist. (Default value = False)

    """

    FN_STATE_POINT = "signac_statepoint.json"
    """The job's state point filename.

    The job state point is a human-readable file containing the job's state
    point that is stored in each job's workspace directory.
    """

    FN_DOCUMENT = "signac_job_document.json"
    "The job's document filename."

    KEY_DATA = "signac_data"
    "The job's datastore key."

    def __init__(self, project, statepoint=None, id_=None, directory_known=False):
        self._project = project
        self._lock = RLock()
        self._initialize_lazy_properties()
        self._directory_known = directory_known

        if statepoint is None and id_ is None:
            raise ValueError("Either statepoint or id_ must be provided.")
        elif statepoint is not None:
            try:
                self._id = calc_id(statepoint) if id_ is None else id_
            except TypeError:
                raise KeyTypeError

            self._cached_statepoint = statepoint
            self._statepoint_requires_init = True
        else:
            # Only an id was provided. State point will be loaded lazily.
            self._id = id_
            self._statepoint_requires_init = True

            # Fetch the cached statepoint from the project's cache. Don't load it
            # from disk on a cache miss (will be loaded on demand).
            try:
                self._cached_statepoint = project._sp_cache[id_]
            except KeyError:
                self._cached_statepoint = None

    def _initialize_lazy_properties(self):
        """Initialize all properties that are designed to be loaded lazily."""
        with self._lock:
            self._path = None
            self._document = None
            self._stores = None
            self._cwd = []

    @property
    def id(self):
        """Get the unique identifier for the job's state point.

        Returns
        -------
        str
            The job id.

        """
        return self._id

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented
        return self.id == other.id and os.path.realpath(self.path) == os.path.realpath(
            other.path
        )

    def __str__(self):
        """Return the job's id."""
        return str(self.id)

    def __repr__(self):
        return "{}(project={}, statepoint={})".format(
            self.__class__.__name__, repr(self._project), self.cached_statepoint
        )

    @property
    def _statepoint_filename(self):
        """Get the path of the state point file for this job."""
        # Performance-critical path. We can rely on the job workspace and state
        # point file name to be well-formed, so just use str.join with os.sep
        # instead of os.path.join for speed.
        return os.sep.join((self.path, self.FN_STATE_POINT))

    @property
    def path(self):
        """str: The path to the job directory.

        See :ref:`signac job -w <signac-cli-job>` for the command line equivalent.
        """
        if self._path is None:
            # We can rely on the project workspace to be well-formed, so just
            # use str.join with os.sep instead of os.path.join for speed.
            self._path = os.sep.join((self._project.workspace, self.id))
        return self._path

    def update_statepoint(self, update, overwrite=False):
        """Change the state point of this job while preserving job data.

        By default, this method will not change existing parameters of the
        state point of the job.

        This method will change the job id if the state point has been altered.

        For more information, see
        `Modifying the State Point <https://signac.readthedocs.io/en/latest/jobs.html#modifying-the-state-point>`_.

        .. warning::

            While appending to a job's state point is generally safe, modifying
            existing parameters may lead to data inconsistency. Use the
            ``overwrite`` argument with caution!

        Parameters
        ----------
        update : dict
            A mapping used for the state point update.
        overwrite : bool, optional
            If False, an error will be raised if the update modifies the values
            of existing keys in the state point. If True, any existing keys will
            be overwritten in the same way as :meth:`dict.update`. Use with
            caution! (Default value = False).

        Raises
        ------
        KeyError
            If the update contains keys which are already part of the job's
            state point and ``overwrite`` is False.
        :class:`~signac.errors.DestinationExistsError`
            If a job associated with the new state point is already initialized.
        OSError
            If the move failed due to an unknown system related error.

        """  # noqa: E501
        statepoint = self.statepoint()
        if not overwrite:
            for key, value in update.items():
                if statepoint.get(key, value) != value:
                    raise KeyError(
                        f"Key {key} was provided but already exists in the "
                        "mapping with another value."
                    )
        statepoint.update(update)
        self.statepoint = statepoint

    @property
    def cached_statepoint(self):
        """Get a copy of the job's state point as a read-only mapping.

        :py:attr:`cached_statepoint` uses the state point cache to provide fast access to
        the job's state point for reading.

        .. note::

            Create and update the state point cache by calling
            :py:meth:`project.update_cache <signac.Project.update_cache>`
            or running ``signac update-cache`` on the command line.

        .. seealso::

            Use :py:attr:`statepoint` to modify the job's state point.

        Returns
        -------
        Mapping
            Returns the job's state point.
        """
        if self._cached_statepoint is None:
            self._cached_statepoint = self._project._get_statepoint(self._id)

        return MappingProxyType(self._cached_statepoint)

    @property
    def statepoint(self):
        """Get or set the job's state point.

        Setting the state point to a different value will change the job id.

        For more information, see
        `Modifying the State Point
        <https://signac.readthedocs.io/en/latest/jobs.html#modifying-the-state-point>`_.

        .. tip::

            Use :py:attr:`cached_statepoint` for fast read-only access to the
            state point.

        .. warning::

            The state point object behaves like a dictionary in most cases,
            but because it persists changes to the filesystem, making a copy
            requires explicitly converting it to a dict. If you need a
            modifiable copy that will not modify the underlying JSON file,
            you can access a dict copy of the state point by calling it, e.g.
            ``sp_dict = job.statepoint()`` instead of ``sp = job.statepoint``.
            For more information, see
            :class:`~synced_collections.backends.collection_json.JSONAttrDict`.

        See :ref:`signac statepoint <signac-cli-statepoint>` for the command line equivalent.

        .. danger::

            Use this function with caution! Resetting a job's state point
            may sometimes be necessary, but can possibly lead to incoherent
            data spaces.

        Returns
        -------
        MutableMapping
            Returns the job's state point. Supports attribute-based access to
            dict keys.
        """
        with self._lock:
            if self._statepoint_requires_init:
                if self._cached_statepoint is None:
                    # Load state point data lazily (on access).
                    self._statepoint = _StatePointDict(
                        jobs=[self],
                        filename=self._statepoint_filename,
                    )
                    statepoint = self._statepoint.load(self.id)

                    # Update the project's state point cache when loaded lazily
                    self._project._register(self.id, statepoint)
                    self._cached_statepoint = statepoint
                else:
                    # Create _StatePointDict lazily with a known statepoint dict.
                    self._statepoint = _StatePointDict(
                        jobs=[self],
                        filename=self._statepoint_filename,
                        data=self._cached_statepoint,
                    )

                self._statepoint_requires_init = False

        return self._statepoint

    @statepoint.setter
    def statepoint(self, new_statepoint):
        """Assign a new state point to this job.

        Parameters
        ----------
        new_statepoint : dict
            The new state point to be assigned.
        """
        with self._lock:
            if self._statepoint_requires_init:
                # Instantiate state point data lazily - no load is required, since
                # we are provided with the new state point data.
                self._statepoint = _StatePointDict(
                    jobs=[self], filename=self._statepoint_filename
                )
                self._statepoint_requires_init = False
            self.statepoint.reset(new_statepoint)

        self._project._register(self.id, new_statepoint)

    @property
    def sp(self):
        """Alias for :attr:`~Job.statepoint`."""
        return self.statepoint

    @sp.setter
    def sp(self, new_sp):
        """Alias for :attr:`~Job.statepoint`."""
        self.statepoint = new_sp

    @property
    def document(self):
        """Get document associated with this job.

        .. warning::

            Even deep copies of :attr:`~Job.document` will modify the same file,
            so changes will still effectively be persisted between deep copies.
            If you need a deep copy that will not modify the underlying
            persistent JSON file, use the call operator to get an equivalent
            plain dictionary: ``job.document()``.
            For more information, see
            :class:`~signac.JSONDict`.

        See :ref:`signac document <signac-cli-document>` for the command line equivalent.

        Returns
        -------
        MutableMapping
            The job document handle. Supports attribute-based access to dict keys.

        """
        with self._lock:
            if self._document is None:
                self.init(validate_statepoint=False)
                fn_doc = os.path.join(self.path, self.FN_DOCUMENT)
                self._document = BufferedJSONAttrDict(
                    filename=fn_doc, write_concern=True
                )
        return self._document

    @document.setter
    def document(self, new_doc):
        """Assign new document data to this job.

        Parameters
        ----------
        new_doc : dict
            The job document handle.

        """
        self.document.reset(new_doc)

    @property
    def doc(self):
        """Alias for :attr:`~Job.document`.

        .. warning::

            Even deep copies of :attr:`~Job.doc` will modify the same file, so
            changes will still effectively be persisted between deep copies.
            If you need a deep copy that will not modify the underlying
            persistent JSON file, use the call operator to get an equivalent
            plain dictionary: ``job.doc()``.

        See :ref:`signac document <signac-cli-document>` for the command line equivalent.

        Returns
        -------
        MutableMapping
            The job document handle. Supports attribute-based access to dict keys.

        """
        return self.document

    @doc.setter
    def doc(self, new_doc):
        """Alias for :attr:`~Job.document`."""
        self.document = new_doc

    @property
    def stores(self):
        """Get HDF5 stores associated with this job.

        Use this property to access an HDF5 file within the job's workspace
        directory using the :class:`~signac.H5Store` dict-like interface.

        This is an example for accessing an HDF5 file called 'my_data.h5' within
        the job's workspace:

        .. code-block:: python

            job.stores['my_data']['array'] = np.random((32, 4))

        This is equivalent to:

        .. code-block:: python

            H5Store(job.fn('my_data.h5'))['array'] = np.random((32, 4))

        Both the :attr:`~job.stores` and the :class:`~signac.H5Store` itself support attribute
        access. The above example could therefore also be expressed as:

        .. code-block:: python

            job.stores.my_data.array = np.random((32, 4))

        Returns
        -------
        :class:`~signac.H5StoreManager`
            The HDF5-Store manager for this job.

        """
        with self._lock:
            if self._stores is None:
                self.init(validate_statepoint=False)
                self._stores = H5StoreManager(self.path)
        return self._stores

    @property
    def data(self):
        """Get data associated with this job.

        This property should be used for large array-like data, which can't be
        stored efficiently in the job document. For examples and usage, see
        `Job Data Storage <https://signac.readthedocs.io/en/latest/jobs.html#job-data-storage>`_.

        Equivalent to:

        .. code-block:: python

                return job.stores['signac_data']

        Returns
        -------
        :class:`~signac.H5Store`
            An HDF5-backed datastore.

        """
        return self.stores[self.KEY_DATA]

    @data.setter
    def data(self, new_data):
        """Assign new data to this job.

        Parameters
        ----------
        new_data : :class:`~signac.H5Store`
            An HDF5-backed datastore.

        """
        self.stores[self.KEY_DATA] = new_data

    @property
    def project(self):
        """Get the project that contains this job.

        Returns
        -------
        signac.Project
            Returns the project containing this job.
        """
        return self._project

    def init(self, force=False, validate_statepoint=True):
        """Initialize the job's workspace directory.

        This function will do nothing if the directory and the job state point
        already exist and the state point is valid.

        Returns the calling job.

        See :ref:`signac job -c <signac-cli-job>` for the command line equivalent.

        Parameters
        ----------
        force : bool, optional
            Overwrite any existing state point files, e.g., to repair them if
            they got corrupted (Default value = False).

        validate_statepoint : bool, optional
            When True (the default), load the job state point and ensure that it matches
            the id. When False, exit early when the job directory exists.

        Returns
        -------
        Job
            The job handle.

        Raises
        ------
        OSError
            If the workspace directory cannot be created or any other I/O error
            occurs when attempting to save the state point file.
        JobsCorruptedError
            If the job state point on disk is corrupted.
        """
        with self._lock:
            try:
                # Fast early exit when not validating.
                if not validate_statepoint:
                    if self._directory_known:
                        return self

                    if os.path.isdir(self.path):
                        self._directory_known = True
                        return self

                # Attempt early exit if the state point file exists and is valid.
                try:
                    statepoint = self.statepoint.load(self.id)
                except Exception:
                    # Any exception means this method cannot exit early.

                    # Create the workspace directory if it does not exist.
                    try:
                        _mkdir_p(self.path)
                    except OSError:
                        logger.error(
                            "Error occurred while trying to create "
                            "workspace directory for job '{}'.".format(self.id)
                        )
                        raise

                    self._directory_known = True

                    # The state point save will not overwrite an existing file on
                    # disk unless force is True, so the subsequent load will catch
                    # when a preexisting invalid file was present.
                    self.statepoint.save(force=force)
                    statepoint = self.statepoint.load(self.id)

                    # Update the project's state point cache if the saved file is valid.
                    self._project._register(self.id, statepoint)
            except Exception:
                logger.error(
                    f"State point file of job '{self.id}' appears to be corrupted."
                )
                raise
        return self

    def clear(self):
        """Remove all job data, but not the job itself.

        This function will do nothing if the job was not previously
        initialized.

        See :ref:`signac rm -c <signac-cli-rm>` for the command line equivalent.

        """
        try:
            for fn in os.listdir(self.path):
                if fn in (self.FN_STATE_POINT, self.FN_DOCUMENT):
                    continue
                path = os.path.join(self.path, fn)
                if os.path.isfile(path):
                    os.remove(path)
                elif os.path.isdir(path):
                    shutil.rmtree(path)
            self.document.clear()
        except OSError as error:
            if error.errno != errno.ENOENT:
                raise error

    def reset(self):
        """Remove all job data, but not the job itself.

        This function will initialize the job if it was not previously
        initialized.

        """
        self.clear()
        self.init()

    def remove(self):
        """Remove the job's workspace including the job document.

        This function will do nothing if the workspace directory
        does not exist.

        See :ref:`signac rm <signac-cli-rm>` for the command line equivalent.

        """
        with self._lock:
            try:
                shutil.rmtree(self.path)
            except OSError as error:
                if error.errno != errno.ENOENT:
                    raise
            else:
                if self._document is not None:
                    try:
                        self._document.clear()
                    except OSError as error:
                        if error.errno != errno.ENOENT:
                            raise error
                    self._document = None
                self._stores = None

            self._directory_known = False

    def move(self, project):
        """Move this job to project.

        This function will attempt to move this instance of job from
        its original project to a different project.

        See :ref:`signac move <signac-cli-move>` for the command line equivalent.

        Parameters
        ----------
        project : :class:`~signac.Project`
            The project to move this job to.

        """
        with self._lock:
            statepoint = self.statepoint()
            dst = project.open_job(statepoint)
            _mkdir_p(project.workspace)
            try:
                os.replace(self.path, dst.path)
            except OSError as error:
                if error.errno == errno.ENOENT:
                    raise RuntimeError(
                        f"Cannot move job '{self}', because it is not initialized!"
                    )
                elif error.errno in (errno.EEXIST, errno.ENOTEMPTY, errno.EACCES):
                    raise DestinationExistsError(dst)
                elif error.errno == errno.EXDEV:
                    raise RuntimeError(
                        "Cannot move jobs across different devices (file systems)."
                    )
                else:
                    raise error
            self.__dict__.update(dst.__dict__)

            # Update the destination project's state point cache
            project._register(self.id, statepoint)

    def sync(self, other, strategy=None, exclude=None, doc_sync=None, **kwargs):
        r"""Perform a one-way synchronization of this job with the other job.

        By default, this method will synchronize all files and document data with
        the other job to this job until a synchronization conflict occurs. There
        are two different kinds of synchronization conflicts:

            1. The two jobs have files with the same, but different content.
            2. The two jobs have documents that share keys, but those keys are
               associated with different values.

        A file conflict can be resolved by providing a 'FileSync' *strategy* or by
        *excluding* files from the synchronization. An unresolvable conflict is indicated with
        the raise of a :class:`~signac.errors.FileSyncConflict` exception.

        A document synchronization conflict can be resolved by providing a doc_sync function
        that takes the source and the destination document as first and second argument.

        Parameters
        ----------
        other : Job
            The other job to synchronize from.
        strategy : callable, optional
            A synchronization strategy for file conflicts. If no strategy is provided, a
            :class:`~signac.errors.SyncConflict` exception will be raised upon conflict
            (Default value = None).
        exclude : str, optional
            A filename exclude pattern. All files matching this pattern will be
            excluded from synchronization (Default value = None).
        doc_sync : attribute or callable from :py:class:`~signac.sync.DocSync`, optional
            A synchronization strategy for document keys. If this argument is None, by default
            no keys will be synchronized upon conflict (Default value = None).
        dry_run : bool, optional
            If True, do not actually perform the synchronization.
        \*\*kwargs :
            Extra keyword arguments will be forwarded to the :meth:`~signac.sync.sync_jobs`
            function which actually executes the synchronization operation.

        Raises
        ------
        :class:`~signac.errors.FileSyncConflict`
            In case that a file synchronization results in a conflict.

        """
        sync_jobs(
            src=other,
            dst=self,
            strategy=strategy,
            exclude=exclude,
            doc_sync=doc_sync,
            **kwargs,
        )

    def fn(self, filename):
        """Prepend a filename with the job path.

        Parameters
        ----------
        filename : str
            The name of the file.

        Returns
        -------
        str
            The absolute path to the file.

        """
        return os.path.join(self.path, filename)

    def isfile(self, filename):
        """Check if a filename exists in the job directory.

        Parameters
        ----------
        filename : str
            The name of the file.

        Returns
        -------
        bool
            True if filename exists in the job directory.

        """
        return os.path.isfile(self.fn(filename))

    def open(self):
        """Enter the job's workspace directory.

        You can use the `Job` class as context manager:

        .. code-block:: python

            with project.open_job(my_statepoint) as job:
                # Manipulate your job data
                pass

        Opening the context will switch into the job's workspace,
        leaving it will switch back to the previous working directory.

        """
        self._cwd.append(os.getcwd())
        self.init(validate_statepoint=False)
        logger.info(f"Enter workspace '{self.path}'.")
        os.chdir(self.path)

    def close(self):
        """Close the job and switch to the previous working directory."""
        try:
            os.chdir(self._cwd.pop())
            logger.info("Leave workspace.")
        except IndexError:
            pass

    def __enter__(self):
        self.open()
        return self

    def __exit__(self, err_type, err_value, tb):
        self.close()
        return False

    def __getstate__(self):
        state = dict(self.__dict__)
        # Locks are not pickleable and must be removed from the state
        del state["_lock"]
        return state

    def __setstate__(self, state):
        # Locks are not pickleable and must be added back to the state
        state["_lock"] = RLock()
        self.__dict__.update(state)
        # We append to a list of jobs rather than replacing to support
        # transparent id updates between shallow copies of a job.
        self.statepoint._jobs.append(self)

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        state = dict(self.__dict__)
        # Locks are not pickleable and must be removed/added back
        del state["_lock"]
        for key, value in state.items():
            setattr(result, key, deepcopy(value, memo))
        result._lock = RLock()
        return result



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import os

import signac


class TestTemporaryProject:
    def test_init_context_manager_constructor(self):
        with signac.TemporaryProject() as tmp_project:
            assert os.path.isdir(tmp_project.path)
            for i in range(10):
                tmp_project.open_job(dict(a=i)).init()
            assert len(tmp_project) == 10
        assert not os.path.isdir(tmp_project.path)

    def test_init_project_method(self):
        with signac.TemporaryProject() as project:
            with project.temporary_project() as tmp_project:
                assert os.path.isdir(tmp_project.path)
                for i in range(10):
                    tmp_project.open_job(dict(a=i)).init()
                assert len(tmp_project) == 10
            assert not os.path.isdir(tmp_project.path)



# Copyright (c) 2022 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
"""Migrate from schema version 1 to version 2.

This migration involves the following changes:
    - Moving the signac.rc config file to .signac/config
    - Moving .signac_shell_history to .signac/shell_history
    - Moving .signac_sp_cache.json.gz to .signac/statepoint_cache.json.gz
    - Removing the project name from the config. Projects are now identified
      solely by their directories.
    - Removing the workspace_dir key from the config. The workspace directory
      is no longer configurable.
"""

import os

from synced_collections.backends.collection_json import BufferedJSONAttrDict

from .._config import _get_project_config_fn
from .._vendor import configobj
from ..project import Project
from .v0_to_v1 import _load_config_v1

# A minimal v2 config.
_CFG = """
schema_version = string(default='0')
"""


def _load_config_v2(root_directory):
    config_fn = os.path.join(root_directory, ".signac", "config")
    if not os.path.isfile(config_fn):
        raise RuntimeError(
            f"The directory {root_directory} does not contain a config file."
        )
    cfg = configobj.ConfigObj(config_fn, configspec=_CFG.split("\n"))
    validator = configobj.validate.Validator()
    if cfg.validate(validator) is not True:
        raise RuntimeError(
            "This project's config file is not compatible with signac's v2 schema."
        )
    return cfg


def _migrate_v1_to_v2(root_directory):
    """Migrate from schema version 1 to version 2."""
    # Load the v1 config.
    cfg = _load_config_v1(root_directory)

    # Try to migrate a custom workspace directory if one exists.
    current_workspace_name = cfg.get("workspace_dir")
    if current_workspace_name is not None:
        if current_workspace_name != "workspace":
            current_workspace = os.path.join(root_directory, current_workspace_name)
            new_workspace = os.path.join(root_directory, "workspace")
            if os.path.exists(new_workspace):
                raise RuntimeError(
                    "Workspace directories are no longer configurable in schema version 2, and "
                    f"must be 'workspace', but {new_workspace} already exists. Please remove or "
                    f"move it so that the currently configured workspace directory "
                    f"{current_workspace} can be moved to {new_workspace}."
                )
            os.replace(current_workspace, new_workspace)
        del cfg["workspace_dir"]

    # Delete project name from config and store in project doc if non-default.
    # For default names, no modifications to the project document should be made.
    if cfg["project"] != "None":
        fn_doc = os.path.join(root_directory, Project.FN_DOCUMENT)
        doc = BufferedJSONAttrDict(filename=fn_doc, write_concern=True)
        doc["signac_project_name"] = cfg["project"]
    del cfg["project"]
    cfg.write()

    # Move signac.rc to .signac/config
    v1_fn = os.path.join(root_directory, "signac.rc")
    v2_fn = _get_project_config_fn(root_directory)
    os.mkdir(os.path.dirname(v2_fn))
    os.replace(v1_fn, v2_fn)

    # Now move all other files.
    files_to_move = {
        ".signac_shell_history": os.path.join(".signac", "shell_history"),
        ".signac_sp_cache.json.gz": os.path.join(".signac", "statepoint_cache.json.gz"),
    }
    for src, dst in files_to_move.items():
        src = os.path.join(root_directory, src)
        if os.path.isfile(src):
            os.replace(src, os.path.join(root_directory, dst))



# Copyright (c) 2018 The Regents of the University of Michigan
# All rights reserved.
# This software is licensed under the BSD 3-Clause License.
import pytest
from synced_collections.numpy_utils import NumpyConversionWarning
from test_project import TestProjectBase

try:
    import numpy  # noqa
    import numpy.testing

    NUMPY = True
except ImportError:
    NUMPY = False


@pytest.mark.skipif(not NUMPY, reason="test requires the numpy package")
class TestNumpyIntegration(TestProjectBase):
    def test_store_number_in_sp_and_doc(self):
        for i in range(10):
            a = numpy.float32(i) if i % 2 else numpy.float64(i)
            b = numpy.float64(i) if i % 2 else numpy.float32(i)
            with pytest.warns(NumpyConversionWarning):
                job = self.project.open_job(dict(a=a))
                job.statepoint
            with pytest.warns(NumpyConversionWarning):
                job.doc.b = b
            numpy.testing.assert_equal(job.doc.b, b)
        for i, job in enumerate(sorted(self.project, key=lambda job: job.sp.a)):
            assert job.sp.a == i
            assert job.doc.b == i

    def test_store_array_in_sp(self):
        for i in range(10):
            with pytest.warns(NumpyConversionWarning):
                self.project.open_job(dict(a=numpy.array([i]))).init()
        for i, job in enumerate(sorted(self.project, key=lambda job: job.sp.a)):
            assert [i] == job.sp.a
            assert numpy.array([i]) == job.sp.a

    def test_store_array_in_doc(self):
        for i in range(10):
            job = self.project.open_job(dict(a=i))
            with pytest.warns(NumpyConversionWarning):
                job.doc.array = numpy.ones(3) * i
            numpy.testing.assert_equal(job.doc.array, numpy.ones(3) * i)
        for i, job in enumerate(sorted(self.project, key=lambda job: job.sp.a)):
            assert i == job.sp.a
            assert (numpy.array([i, i, i]) == job.doc.array).all()
            assert [i] * 3 == job.doc.array

    def test_store_zero_dim_array_in_sp(self):
        # Zero-dimensional arrays have size 1, and their tolist() method
        # returns a single value.
        value = 1.0
        with pytest.warns(NumpyConversionWarning):
            job = self.project.open_job(dict(a=numpy.array(value))).init()
        assert value == job.sp.a
        assert numpy.array(value) == job.sp.a

    def test_store_zero_dim_array_in_doc(self):
        # Zero-dimensional arrays have size 1, and their tolist() method
        # returns a single value.
        value = 1.0
        job = self.project.open_job(dict(a=1)).init()
        with pytest.warns(NumpyConversionWarning):
            job.doc.array = numpy.array(value)
        numpy.testing.assert_equal(job.doc.array, numpy.array(value))
        assert value == job.doc.array
        assert numpy.array(value) == job.doc.array
